////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <gtf/scdr/Allocator.h>
#include <gtf/scdr/FramebufferClearFlags.h>
#include <gtf/scdr/SDK.h>
#include <gtf/scdr/RenderObject.h>

#include "CrepuscularRenderer.h"
#include "DirectX11Shader.h"
#include "OpenGLES3Shader.h"
#include "PreparedStageImpl.h"

class CrepuscularStackFrame : public gtf::scdr::RenderObjectStackFrame2D
{
public:
    virtual void load(gtf::scdr::RenderObjectStackFrame2DValues& valuesOut_) const
    {
        valuesOut_.coloration         = &coloration;
        valuesOut_.framebufferId      = framebufferId;
        valuesOut_.framebuffer        = framebuffer;
        valuesOut_.framebuffer2Screen = &framebuffer2Screen;
        valuesOut_.scissor            = &scissor;
        valuesOut_.viewport           = &viewport;
        valuesOut_.viewProjMat        = &viewProjMat;
        valuesOut_.widget2Framebuffer = &widget2Framebuffer;
        valuesOut_.depth              = depth;
    }

public:
    uint32_t                framebufferId;
    gtf::scdr::Framebuffer* framebuffer;
    gtf::math::float4x4     framebuffer2Screen;
    gtf::math::float4x4     viewProjMat;
    gtf::math::float4x4     widget2Framebuffer;
    gtf::scdr::Color        coloration;
    gtf::scdr::Viewport     viewport;
    gtf::scdr::Scissor      scissor;
    uint32_t*               depth;
};

class CrepuscularRenderObject
    : public gtf::scdr::RenderObject
    , public gtf::scdr::TouchObject
{
private:
    CrepuscularRenderObject(const CrepuscularRenderObject&);
    CrepuscularRenderObject& operator=(const CrepuscularRenderObject&);

public:
    CrepuscularRenderObject(const CrepuscularGraphicObjects& graphicObjects_, const gtf::widgetset::WidgetModel& widget_, const gtf::scdr::WidgetFeaturesCommon2D& features_)
        : next(NULL)
        , graphicObjects(graphicObjects_)
        , constantBuffer(NULL)
        , framebuffer(NULL)
        , offscreenFramebuffer(NULL)
        , viewport(gtf::scdr::CViewport(0, 0, 0, 0))
        , width(0)
        , height(0)
        , zid(0)
        , fbid(0)
        , offscreenFbId(0)
    {
        gtf::scdr::GetWidgetFeaturesCommonModels2D(widget_, features_, features);
    }

public: // RenderObject interface
    struct StackFrameDepthPair
    {
        CrepuscularStackFrame frame;
        uint32_t depth;
    };

    virtual gtf::scdr::RenderObjectStackFrame* createStackFrame(gtf::scdr::RenderObjectStackFrame& parent_, gtf::scdr::Allocator& allocator_)
    {
        GTF_UNUSED_PARAM(parent_);
        StackFrameDepthPair* const pair = ALLOC_OBJECT(allocator_, StackFrameDepthPair)();

        if (pair)
        {
            pair->frame.depth = &pair->depth;
            return &pair->frame;
        }
        else
        {
            return NULL;
        }
    }

    virtual void destroyStackFrame(gtf::scdr::RenderObjectStackFrame* frame_, gtf::scdr::Allocator& allocator_)
    {
        void* const crepuscularFrame = static_cast<void*>(frame_);

        if (crepuscularFrame != NULL)
        {
            StackFrameDepthPair* pair = static_cast<StackFrameDepthPair*>(crepuscularFrame);
            FREE_OBJECT(allocator_, pair);
        }
    }

    virtual bool transformCoordinates(gtf::scdr::TransformMode::Value mode_, uint32_t coordCount_, gtf::scdr::TransformCoordinate* coordInOut_) const
    {
        switch (mode_)
        {
        case gtf::scdr::TransformMode::Widget2Screen:
            for (uint32_t i = 0; i < coordCount_; ++i)
            {
                coordInOut_[i].x /= static_cast<float>(width);
                coordInOut_[i].y /= static_cast<float>(height);
                gtf::scdr::CalculateCoordinateTransformation(unit2Screen, coordInOut_[i]);
            }
            return true;

        case gtf::scdr::TransformMode::Screen2Widget:
        {
            gtf::math::float4x4 invT;
            if (unit2Screen.inverse(invT))
            {
                for (uint32_t i = 0; i < coordCount_; ++i)
                {
                    gtf::scdr::CalculateCoordinateTransformation(invT, coordInOut_[i]);
                    coordInOut_[i].x *= static_cast<float>(width);
                    coordInOut_[i].y *= static_cast<float>(height);
                }
                return true;
            }
            return false;
        }

        case gtf::scdr::TransformMode::Unit2Screen:
            for (uint32_t i = 0; i < coordCount_; ++i)
            {
                gtf::scdr::CalculateCoordinateTransformation(unit2Screen, coordInOut_[i]);
            }
            return true;

        case gtf::scdr::TransformMode::Screen2Unit:
        {
            gtf::math::float4x4 invT;
            if (unit2Screen.inverse(invT))
            {
                for (uint32_t i = 0; i < coordCount_; ++i)
                {
                    gtf::scdr::CalculateCoordinateTransformation(invT, coordInOut_[i]);
                }
                return true;
            }
            return false;
        }
        default:
            return false;
        }
    }

    virtual const gtf::scdr::TouchObject* getTouchObject()
    {
        if (features.touch)
        {
            return this;
        }
        return NULL;
    }

public: // TouchObject interface
    virtual void getBoundingRectangle(gtf::scdr::Rectangle<int16_t>& rectOut_) const
    {
        gtf::scdr::CalculateBoundingRectangle(*this, rectOut_);
    }

    virtual bool intersect(const gtf::scdr::ScreenPosition& pos_) const
    {
        return gtf::scdr::IntersectsBoundingRectangle(*this, pos_);
    }

public:
    CrepuscularRenderObject*                next;
    gtf::scdr::WidgetFeaturesCommon2DModels features;

    const CrepuscularGraphicObjects&        graphicObjects;

    CrepuscularConstantBufferBlock          constantBufferBlock;
    gtf::scdr::ConstantBuffer*              constantBuffer;
    gtf::scdr::Framebuffer*                 framebuffer;
    gtf::scdr::Framebuffer*                 offscreenFramebuffer;
    gtf::scdr::Viewport                     viewport;
    gtf::math::float4x4                     unit2Screen;

    uint32_t                                width;
    uint32_t                                height;
    uint32_t                                zid;
    uint32_t                                fbid;
    uint32_t                                offscreenFbId;
};

static bool ClearOffscreenFrambuffer(void* effect_, gtf::scdr::GraphicsExecutionContext& context_)
{
    static const uint32_t clearFlags =
        gtf::scdr::FramebufferClearFlags::Color |
        gtf::scdr::FramebufferClearFlags::Depth |
        gtf::scdr::FramebufferClearFlags::Stencil;

    static const float clearColor[4] =
    {
        0, 0, 0, 0
    };

    if (effect_)
    {
        CrepuscularRenderObject* const object = static_cast<CrepuscularRenderObject*>(effect_);
        context_.clearFramebuffer(object->offscreenFramebuffer, clearFlags, clearColor, 1.0, 0);

        return true;
    }
    return false;
}

static bool DrawEffect(void* crepuscularRenderObject_, gtf::scdr::GraphicsExecutionContext& context_)
{
    GTF_UNUSED_PARAM(context_);

    CrepuscularRenderObject* const object = static_cast<CrepuscularRenderObject*>(crepuscularRenderObject_);

    if (object->framebuffer)
    {
        CrepuscularShaderResources resources;

        const CrepuscularGraphicObjects& graphicObjects = object->graphicObjects;

        gtf::scdr::Texture2D* const texture = context_.getTextureAttachment(
            object->offscreenFramebuffer,
            gtf::scdr::FramebufferTextureAttachment::Color0);

        gtf::scdr::SetResourceSlot(resources.constantBuffer, object->constantBuffer);
        gtf::scdr::SetResourceSlot(resources.colorTexture  , texture);
        gtf::scdr::SetResourceSlot(resources.colorSampler  , graphicObjects.colorSampler);

        const uint32_t offset = 0;
        const uint32_t stride = 4 * sizeof(float);

        context_.setFramebuffer(object->framebuffer);
        context_.setViewport(object->viewport);

        context_.setVertexBuffers(1, &graphicObjects.vb, &stride, &offset);
        context_.setPipelineStateObject(graphicObjects.pso);
        context_.setPipelineResources(graphicObjects.rbt, &resources.constantBuffer, sizeof(CrepuscularShaderResources) / sizeof(resources.constantBuffer));

        context_.draw(4, 0);
    }

    return true;
}

static bool UpdateEffect(void* crepuscularRenderObject_, gtf::scdr::GraphicsExecutionContext& context_)
{
    const CrepuscularRenderObject* object = static_cast<const CrepuscularRenderObject*>(crepuscularRenderObject_);

    return context_.updateConstantBuffer(object->constantBuffer, &object->constantBufferBlock, sizeof(object->constantBufferBlock));
}

struct CrepuscularVertex
{
    float pos[2];
    float tex[2];
};

static const CrepuscularVertex CrepuscularQuad[] =
{
    { 0.f, 0.f, 0.f, 1.f },
    { 0.f, 1.f, 0.f, 0.f },
    { 1.f, 0.f, 1.f, 1.f },
    { 1.f, 1.f, 1.f, 0.f }
};

static bool CreateCrepuscularQuad(gtf::scdr::GraphicsContext& context_, gtf::scdr::VertexBuffer*& vbOut_)
{
    vbOut_ = context_.createVertexBuffer(
        gtf::scdr::CBufferDescriptor(gtf::scdr::BufferUsage::Immutable, sizeof(CrepuscularQuad)),
        CrepuscularQuad);

    return vbOut_ != NULL;
}

static bool CreateCrepuscularShaderProgram(gtf::scdr::GraphicsApi::Value driver_, gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram*& progOut_)
{
    gtf::scdr::ProgramDescriptor pd;
    memset(&pd, 0, sizeof(pd));

    switch (driver_)
    {
    case gtf::scdr::GraphicsApi::DirectX11:
        pd.vs.source       = VertexShaderD3D11;
        pd.vs.sourceLength = strlen(VertexShaderD3D11);
        pd.ps.source       = PixelShaderD3D11;
        pd.ps.sourceLength = strlen(PixelShaderD3D11);
        break;

    case gtf::scdr::GraphicsApi::OpenGLES3:
        pd.vs.source       = VertexShaderGLES3;
        pd.vs.sourceLength = strlen(VertexShaderGLES3);
        pd.ps.source       = FragmentShaderGLES3;
        pd.ps.sourceLength = strlen(FragmentShaderGLES3);
        break;

    default:
        return false;
    }

    progOut_ = context_.createShaderProgram(pd);

    return progOut_ != NULL;
}

static bool CreateCrepuscularPso(gtf::scdr::ShaderProgram* prog_, gtf::scdr::GraphicsContext& context_, gtf::scdr::PipelineStateObject*& psoOut_)
{
    const gtf::scdr::PipelineInputDescriptor pid[] =
    {
        { "POSITION", 0, gtf::scdr::Format::Float2, 0, 0, gtf::scdr::InputClassification::VertexData, 0 },
        { "TEXCOORD", 0, gtf::scdr::Format::Float2, 0, 8, gtf::scdr::InputClassification::VertexData, 0 }
    };

    gtf::scdr::PipelineStateObjectDescriptor psod;
    psod.inputDescriptorsCount = ARRAY_SIZE(pid);
    psod.inputDescriptors      = pid;
    psod.program               = prog_;
    psod.primitiveTopology     = gtf::scdr::PrimitiveTopology::TriangleStrip;

    gtf::scdr::CreateAdditiveBlendStateDescriptor(psod.blendStateDescriptor);
    gtf::scdr::CreateDefaultRasterizerStateDescriptor(psod.rasterizerStateDescriptor);
    gtf::scdr::CreateDefaultDepthStencilStateDescriptor(psod.depthStencilStateDescriptor);

    psoOut_ = context_.createPipelineStateObject(psod);
    return (psoOut_ != NULL);
}

static bool CreateCrepuscularSampler(gtf::scdr::GraphicsContext& context_, gtf::scdr::SamplerState*& samplerOut_)
{
    samplerOut_ = context_.createSamplerState(gtf::scdr::CSamplerStateDescriptor(gtf::scdr::TextureFilterMode::Point));

    return samplerOut_ != NULL;
}

static bool CreateCrepuscularResourceBindingTable(
    gtf::scdr::GraphicsApi::Value driver_,
    gtf::scdr::ShaderProgram* prog_,
    gtf::scdr::GraphicsContext& context_,
    gtf::scdr::ResourceBindingTable*& rbtOut_)
{

    //
    // Resource bindings for D3D11
    //
    const gtf::scdr::ResourceBindingDescriptor d3dRbd[] =
    {
        { "cbEffectParams", gtf::scdr::ResourceType::ConstantBuffer, 0, 0 },
        { "colorTexture"  , gtf::scdr::ResourceType::Texture2D     , 1, 0 },
        { "colorSampler"  , gtf::scdr::ResourceType::Sampler       , 2, 0 }
    };

    //
    // Resource bindings for OpenGL|ES 3.x
    //
    const gtf::scdr::ResourceBindingDescriptor gl3Rbd[] =
    {
        { "cbEffectParams", gtf::scdr::ResourceType::ConstantBuffer, 0, 0 },
        { "colorTexture"  , gtf::scdr::ResourceType::Texture2D     , 1, 0 },
        { "colorTexture"  , gtf::scdr::ResourceType::Sampler       , 2, 0 }
    };

    const gtf::scdr::ResourceBindingDescriptor* rbd = NULL;
    uint32_t rbdcount                               = 0;

    switch (driver_)
    {
    case gtf::scdr::GraphicsApi::DirectX11:
        rbd      = d3dRbd;
        rbdcount = ARRAY_SIZE(d3dRbd);
        break;

    case gtf::scdr::GraphicsApi::OpenGLES3:
        rbd      = gl3Rbd;
        rbdcount = ARRAY_SIZE(gl3Rbd);
        break;

    default:
        return false;
    }

    const gtf::scdr::ResourceBindingTableDescriptor rbtd =
    {
        prog_,
        rbd,
        rbdcount
    };

    rbtOut_ = context_.createResourceBindingTable(rbtd);

    return (rbtOut_ != NULL);
}

gtf::scdr::WidgetRenderer* CrepuscularRenderer::createInstance (
    const gtf::viewfactory::ViewFactory& factory_,
    gtf::scdr::GraphicsContext& context_,
    const void* createParams_)
{

    const gtf::scdr::SDKHandle& handle = *static_cast<const gtf::scdr::SDKHandle*>(createParams_);
    gtf::scdr::Allocator& allocator    = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);

    return ALLOC_OBJECT(allocator, CrepuscularRenderer)(factory_, handle->getGraphicsApi());
}

void CrepuscularRenderer::destroyInstance(
    gtf::scdr::WidgetRenderer* renderer_,
    gtf::scdr::GraphicsContext& context_)
{
    gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    FREE_OBJECT(allocator, static_cast<CrepuscularRenderer*>(renderer_));
}

CrepuscularRenderer::CrepuscularRenderer(const gtf::viewfactory::ViewFactory& factory_, gtf::scdr::GraphicsApi::Value driver_)
{
    gtf::scdr::ResolvePropertiesCommon2D(factory_, CREPUSCULAR_PLUGIN_NAME, properties.common2d);
    gtf::scdr::ResolveFeaturesCommon2D(factory_, features.common2d);

    memset(&graphicObjects, 0, sizeof(graphicObjects));
    graphicObjects.driver = driver_;

    gtf::viewfactory::ViewFactory::WidgetPropertyReference wpr[] =
    {
        { { CREPUSCULAR_PLUGIN_NAME, gtf::types::limits::uint16_max }, { "color", gtf::types::limits::uint16_max } },
        { { CREPUSCULAR_PLUGIN_NAME, gtf::types::limits::uint16_max }, { "x-pos"  , gtf::types::limits::uint16_max } },
        { { CREPUSCULAR_PLUGIN_NAME, gtf::types::limits::uint16_max }, { "y-pos"  , gtf::types::limits::uint16_max } },
        { { CREPUSCULAR_PLUGIN_NAME, gtf::types::limits::uint16_max }, { "samples", gtf::types::limits::uint16_max } },
        { { CREPUSCULAR_PLUGIN_NAME, gtf::types::limits::uint16_max }, { "decay"  , gtf::types::limits::uint16_max } },
        { { CREPUSCULAR_PLUGIN_NAME, gtf::types::limits::uint16_max }, { "density", gtf::types::limits::uint16_max } },
        { { CREPUSCULAR_PLUGIN_NAME, gtf::types::limits::uint16_max }, { "weight" , gtf::types::limits::uint16_max } }
    };

    factory_.resolveProperties(wpr, ARRAY_SIZE(wpr));

    properties.color   = wpr[0].widgetPropertyRef.ref;
    properties.xpos    = wpr[1].widgetPropertyRef.ref;
    properties.ypos    = wpr[2].widgetPropertyRef.ref;
    properties.samples = wpr[3].widgetPropertyRef.ref;
    properties.decay   = wpr[4].widgetPropertyRef.ref;
    properties.density = wpr[5].widgetPropertyRef.ref;
    properties.weight  = wpr[6].widgetPropertyRef.ref;
}

CrepuscularRenderer::~CrepuscularRenderer()
{
    GTF_OSAL_ASSERT(graphicObjects.colorSampler == NULL);
    GTF_OSAL_ASSERT(graphicObjects.pso == NULL);
    GTF_OSAL_ASSERT(graphicObjects.rbt == NULL);
    GTF_OSAL_ASSERT(graphicObjects.vb == NULL);
}

bool CrepuscularRenderer::init(gtf::scdr::GraphicsContext& context_)
{
    bool success                      = true;
    gtf::scdr::ShaderProgram* program = NULL;

    success = CreateCrepuscularQuad(context_, this->graphicObjects.vb);
    if (success)
    {
        success = CreateCrepuscularShaderProgram(graphicObjects.driver, context_, program);
    }

    if (success)
    {
        success = CreateCrepuscularPso(program, context_, graphicObjects.pso);
    }

    if (success)
    {
        success = CreateCrepuscularResourceBindingTable(graphicObjects.driver, program, context_, graphicObjects.rbt);
    }

    if (success)
    {
        success = CreateCrepuscularSampler(context_, graphicObjects.colorSampler);
    }

    if (program)
    {
        context_.destroyShaderProgram(program);
    }

    return success;
}

void CrepuscularRenderer::deinit(gtf::scdr::GraphicsContext& context_)
{
    context_.destroyVertexBuffer(graphicObjects.vb);
    context_.destroyPipelineStateObject(graphicObjects.pso);
    context_.destroyResourceBindingTable(graphicObjects.rbt);
    context_.destroySamplerState(graphicObjects.colorSampler);

    graphicObjects.colorSampler = NULL;
    graphicObjects.pso          = NULL;
    graphicObjects.rbt          = NULL;
    graphicObjects.vb           = NULL;
}

bool CrepuscularRenderer::update(gtf::scdr::GraphicsContext& context_)
{
    for (CrepuscularRenderObject* object = objects; object; object = object->next)
    {
        context_.executeUpdateFunction(UpdateEffect, object);
    }
    return true;
}

bool CrepuscularRenderer::render(gtf::scdr::GraphicsContext& context_)
{
    for (CrepuscularRenderObject* object = objects; object; object = object->next)
    {
        context_.executeRenderFunction(
            ClearOffscreenFrambuffer,
            object,
            gtf::scdr::CScheduleDescriptor(object->offscreenFbId, true));

        context_.executeRenderFunction(
            DrawEffect,
            object,
            gtf::scdr::CScheduleDescriptor(graphicObjects.pso, object->fbid, object->zid, false));
    }
    return true;
}

gtf::scdr::RenderObject* CrepuscularRenderer::createRenderObject(
    const gtf::widgetset::WidgetModel& widget_,
    const gtf::coreruntime::CoreContext& widgetContext_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    GTF_UNUSED_PARAM(widgetContext_);

    gtf::scdr::Allocator& allocator       = graphicsContext_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    CrepuscularRenderObject* const object = ALLOC_OBJECT(allocator, CrepuscularRenderObject)(graphicObjects, widget_, features.common2d);

    if (object)
    {
        object->constantBuffer = graphicsContext_.createConstantBuffer(
            gtf::scdr::CBufferDescriptor(gtf::scdr::BufferUsage::Dynamic, sizeof(CrepuscularConstantBufferBlock)),
            NULL);

        if (object->constantBuffer)
        {
            return object;
        }
        else
        {
            FREE_OBJECT(allocator, object);
            return NULL;
        }
    }
    return NULL;
}

void CrepuscularRenderer::destroyRenderObject(
    gtf::scdr::RenderObject* object_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    if (object_)
    {
        gtf::scdr::Allocator& allocator       = graphicsContext_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
        CrepuscularRenderObject* const object = static_cast<CrepuscularRenderObject*>(object_);
        if (object->constantBuffer)
        {
            graphicsContext_.destroyConstantBuffer(object->constantBuffer);
        }

        if (object->offscreenFramebuffer)
        {
            graphicsContext_.destroyFramebuffer(object->offscreenFramebuffer);
        }

        FREE_OBJECT(allocator, object);
    }
}

static uint32_t GetMaxDepth(const gtf::scdr::Framebuffer* fb_)
{
    GTF_OSAL_ASSERT(fb_);

    const gtf::scdr::FramebufferDescriptor& desc = fb_->getDescriptor();
    switch (desc.depthStencilFormat)
    {
    case gtf::scdr::TextureFormat::D16:   return (1U << 16);
    case gtf::scdr::TextureFormat::D32:   return (1U << 24);
    case gtf::scdr::TextureFormat::D24:   return (1U << 24);
    case gtf::scdr::TextureFormat::D24S8: return (1U << 24);
    default:                                   return 0;
    }
}

bool CrepuscularRenderer::traverse(
    gtf::widgetset::WidgetModel& widget_,
    const gtf::coreruntime::CoreContext& widgetContext_,
    const gtf::scdr::RenderObjectStackFrame& parentStackFrame_,
    gtf::scdr::RenderObjectStackFrame& objectStackFrame_,
    gtf::scdr::RenderObject& object_,
    gtf::scdr::TraverserContext& traverserContext_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    CrepuscularStackFrame& objectFrame    = static_cast<CrepuscularStackFrame&>(objectStackFrame_);
    CrepuscularRenderObject& object       = static_cast<CrepuscularRenderObject&>(object_);
    CrepuscularConstantBufferBlock& block = object.constantBufferBlock;

    gtf::scdr::WidgetPropertiesCommon2DValues propertyValues;
    if (!gtf::scdr::GetWidgetPropertiesCommon2DValues(widget_, widgetContext_, properties.common2d, propertyValues))
    {
        return false;
    }

    if (propertyValues.width <= 0 || propertyValues.height <= 0)
    {
        return false;
    }

    gtf::scdr::WidgetFeaturesCommon2DValues featureValues;
    gtf::scdr::GetWidgetFeaturesCommon2DValues(object.features, widgetContext_, features.common2d, featureValues);

    gtf::scdr::RenderObjectStackFrame2DValues stackValues;
    gtf::scdr::GetRenderObjectStackFrameValues2D(parentStackFrame_, stackValues);

    gtf::math::float4x4 widget2Framebuffer = *stackValues.widget2Framebuffer * gtf::math::float4x4::Translation2D(
        static_cast<float>(propertyValues.x), static_cast<float>(propertyValues.y));

    gtf::math::float4x4 featureT;
    if (gtf::scdr::CalculateFeatureTransformMatrix2D(featureValues, featureT))
    {
        widget2Framebuffer *= featureT;
    }

    const uint32_t widgetDepth = gtf::scdr::GetWidgetDepth(stackValues);

    const gtf::math::float4x4 unit2Framebuffer = widget2Framebuffer *
        gtf::math::float4x4::Translation(0.f, 0.f, static_cast<float>(widgetDepth)) *
        gtf::math::float4x4::Scaling2D(static_cast<float>(propertyValues.width), static_cast<float>(propertyValues.height));

    gtf::scdr::LoadValue(widget_, widgetContext_, properties.color  , block.color);
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.decay  , block.decay);
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.density, block.density);
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.samples, block.samples);
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.weight , block.weight);
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.xpos   , block.xpos);
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.ypos   , block.ypos);

    const gtf::scdr::Color coloration = gtf::scdr::CalculateColoration(featureValues, stackValues);
    block.color *= coloration;

    gtf::math::float4x4::store(object.constantBufferBlock.object2Screen, (*stackValues.viewProjMat * unit2Framebuffer).transpose());

    if (object.width != static_cast<uint32_t>(propertyValues.width) || object.height != static_cast<uint32_t>(propertyValues.height))
    {
        if (object.offscreenFramebuffer)
        {
            graphicsContext_.destroyFramebuffer(object.offscreenFramebuffer);
        }

        gtf::scdr::FramebufferDescriptor fbdesc = stackValues.framebuffer->getDescriptor();
        fbdesc.flags                            = 0;
        fbdesc.width                            = static_cast<uint32_t>(propertyValues.width);
        fbdesc.height                           = static_cast<uint32_t>(propertyValues.height);

        object.offscreenFramebuffer = graphicsContext_.createFramebuffer(fbdesc);
        object.width                = propertyValues.width;
        object.height               = propertyValues.height;
    }

    object.zid           = widgetDepth;
    object.fbid          = stackValues.framebufferId;
    object.framebuffer   = stackValues.framebuffer;
    object.viewport      = *stackValues.viewport;
    object.unit2Screen   = *stackValues.framebuffer2Screen * widget2Framebuffer * gtf::math::float4x4::Scaling2D(static_cast<float>(propertyValues.width), static_cast<float>(propertyValues.height));
    object.offscreenFbId = traverserContext_.createFramebufferId();

    const uint32_t maxDepth = GetMaxDepth(object.offscreenFramebuffer);
    if (maxDepth == 0)
    {
        return false;
    }

    objectFrame.viewport           = gtf::scdr::CViewport(0, 0, static_cast<uint32_t>(propertyValues.width), static_cast<uint32_t>(propertyValues.height));
    objectFrame.scissor            = gtf::scdr::DefaultScissor;
    objectFrame.framebuffer2Screen = *stackValues.framebuffer2Screen * widget2Framebuffer;
    objectFrame.coloration         = coloration;
    objectFrame.framebuffer        = object.offscreenFramebuffer;
    objectFrame.framebufferId      = object.offscreenFbId;
    *objectFrame.depth             = maxDepth - 1;

    gtf::scdr::GetViewProjectionMatrix2D(
        propertyValues.width,
        propertyValues.height,
        maxDepth,
        graphicObjects.driver,
        objectFrame.viewProjMat);

    if (featureValues.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Viewport)
    {
        objectFrame.widget2Framebuffer = gtf::math::float4x4::Translation2D(
            static_cast<float>(featureValues.viewportX),
            static_cast<float>(featureValues.viewportY));
    }
    else
    {
        objectFrame.widget2Framebuffer = gtf::math::float4x4::Identity();
    }

    object.next = objects;
    objects     = &object;

    return true;
}

bool CrepuscularRenderer::begin(gtf::scdr::GraphicsContext& context_)
{
    this->objects = NULL;

    GTF_UNUSED_PARAM(context_);
    return true;
}

bool CrepuscularRenderer::end(gtf::scdr::GraphicsContext& context_)
{
    GTF_UNUSED_PARAM(context_);
    return true;
}
