////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include "CustomEffectRenderObject.h"
#include "CustomEffectUtils.h"

CustomEffectRenderObject* CustomEffectRenderObject::create(
    const gtf::widgetset::WidgetModel& widget_,
    const gtf::scdr::WidgetFeaturesCommon2D& features_,
    uint32_t imageCount_,
    uint32_t constantBufferSize_,
    gtf::scdr::GraphicsContext& gc_,
    bool isOffscreen_)
{
    gtf::scdr::Allocator& defaultAllocator = gc_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);

    const uint32_t objsize   = Align(sizeof(CustomEffectRenderObject), sizeof(void*));
    const uint32_t imgsize   = Align(imageCount_ * sizeof(gtf::scdr::ImageResourceHandle), sizeof(void*));
    const uint32_t texsize   = Align(imageCount_ * sizeof(gtf::scdr::Texture2D*), sizeof(void*));
    const uint32_t cbsize    = Align(constantBufferSize_, sizeof(void*));
    const uint32_t allocsize = objsize + imgsize + texsize + cbsize;

    void* mem = defaultAllocator.alloc(allocsize, __FILE__, __LINE__);
    if (mem)
    {
        CustomEffectRenderObject* object = new(mem)CustomEffectRenderObject(widget_, features_);

        if (imageCount_)
        {
            object->imageCount = imageCount_;
            object->images     = reinterpret_cast<gtf::scdr::ImageResourceHandle*>(static_cast<char*>(mem)+objsize);

            for (uint32_t i = 0; i < imageCount_; ++i)
            {
                new(object->images + i)gtf::scdr::ImageResourceHandle();
            }

            object->textures = static_cast<gtf::scdr::Texture2D**>(static_cast<void*>(static_cast<char*>(mem)+objsize + imgsize));
            memset(object->textures, 0, texsize);
        }

        if (cbsize)
        {
            object->buffer = static_cast<char*>(mem)+objsize + imgsize + texsize;
        }
        object->isOffscreen = isOffscreen_;

        return object;
    }
    return NULL;
}

void CustomEffectRenderObject::destroy(CustomEffectRenderObject* object_, gtf::scdr::GraphicsContext& gc_)
{
    gtf::scdr::Allocator& defaultAllocator = gc_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    if (object_)
    {
        for (uint32_t i = 0; i < object_->imageCount; ++i)
        {
            object_->images[i].~ImageResourceHandle();

            gc_.destroyTexture2D(object_->textures[i]);
        }

        if (object_->offscreenFb)
        {
            gc_.destroyFramebuffer(object_->offscreenFb);
        }

        FREE_OBJECT(defaultAllocator, object_);
    }
}

CustomEffectRenderObject::CustomEffectRenderObject(const gtf::widgetset::WidgetModel& widget_, const gtf::scdr::WidgetFeaturesCommon2D& features_)
    : next(NULL)
    , buffer(NULL)
    , images(NULL)
    , fb(NULL)
    , offscreenFb(NULL)
    , textures(NULL)
    , objViewProjMat(gtf::math::float4x4::Identity())
    , unit2Screen(gtf::math::float4x4::Identity())
    , width(0)
    , height(0)
    , fbId(0)
    , offscreenFbId(0)
    , drawId(0)
    , imageCount(0)
    , isOffscreen(false)
{
    gtf::scdr::GetWidgetFeaturesCommonModels2D(widget_, features_, features);
}

struct StackFrameDepthPair
{
    CustomEffectStackFrame frame;
    uint32_t depth;
};

gtf::scdr::RenderObjectStackFrame* CustomEffectRenderObject::createStackFrame(gtf::scdr::RenderObjectStackFrame& parent_, gtf::scdr::Allocator& allocator_)
{
    GTF_UNUSED_PARAM(parent_);
    StackFrameDepthPair* const pair = ALLOC_OBJECT(allocator_, StackFrameDepthPair)();

    if (pair)
    {
        pair->frame.depth = &pair->depth;
        return &pair->frame;
    }
    else
    {
        return NULL;
    }
}

void CustomEffectRenderObject::destroyStackFrame(gtf::scdr::RenderObjectStackFrame* frame_, gtf::scdr::Allocator& allocator_)
{
    void* const customEffectFrame = static_cast<void*>(frame_);

    if (customEffectFrame != NULL)
    {
        StackFrameDepthPair* pair = static_cast<StackFrameDepthPair*>(customEffectFrame);
        FREE_OBJECT(allocator_, pair);
    }
}

bool CustomEffectRenderObject::transformCoordinates(gtf::scdr::TransformMode::Value mode_, uint32_t coordCount_, gtf::scdr::TransformCoordinate* coordInOut_) const
{
    switch (mode_)
    {
    case gtf::scdr::TransformMode::Widget2Screen:
        if (width > 0 && height > 0)
        {
            for (uint32_t i = 0; i < coordCount_; ++i)
            {
                coordInOut_[i].x /= static_cast<float>(width);
                coordInOut_[i].y /= static_cast<float>(height);

                gtf::scdr::CalculateCoordinateTransformation(unit2Screen, coordInOut_[i]);
            }
            return true;
        }
        return false;

    case gtf::scdr::TransformMode::Screen2Widget:
    {
        gtf::math::float4x4 invT;
        if (unit2Screen.inverse(invT))
        {
            for (uint32_t i = 0; i < coordCount_; ++i)
            {
                gtf::scdr::CalculateCoordinateTransformation(invT, coordInOut_[i]);
                coordInOut_[i].x *= static_cast<float>(width);
                coordInOut_[i].y *= static_cast<float>(height);
            }
            return true;
        }
        return false;
    }

    case gtf::scdr::TransformMode::Unit2Screen:
        for (uint32_t i = 0; i < coordCount_; ++i)
        {
            gtf::scdr::CalculateCoordinateTransformation(unit2Screen, coordInOut_[i]);
        }
        return true;

    case gtf::scdr::TransformMode::Screen2Unit:
    {
         gtf::math::float4x4 invT;
         if (unit2Screen.inverse(invT))
         {
             for (uint32_t i = 0; i < coordCount_; ++i)
             {
                 gtf::scdr::CalculateCoordinateTransformation(invT, coordInOut_[i]);
             }
             return true;
         }
         return false;
    }
    default:
        return false;
    }
}

const gtf::scdr::TouchObject* CustomEffectRenderObject::getTouchObject()
{
    if (features.touch && width > 0 && height > 0)
    {
        return this;
    }

    return NULL;
}

void CustomEffectRenderObject::getBoundingRectangle(gtf::scdr::Rectangle<int16_t>& rectOut_) const
{
    gtf::scdr::CalculateBoundingRectangle(*this, rectOut_);
}

bool CustomEffectRenderObject::intersect(const gtf::scdr::ScreenPosition& pos_) const
{
    return gtf::scdr::IntersectsBoundingRectangle(*this, pos_);
}
