////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#ifndef CUSTOM_EFFECT_RENDERER_H_INCLUDED
#define CUSTOM_EFFECT_RENDERER_H_INCLUDED

#include <gtf/resources/ImageResource.h>

#include <gtf/scdr/Framebuffer.h>
#include <gtf/scdr/GraphicsApi.h>
#include <gtf/scdr/GraphicsContext.h>
#include <gtf/scdr/GraphicsExecutionContext.h>
#include <gtf/scdr/RenderObject.h>
#include <gtf/scdr/RenderObjectUtils.h>
#include <gtf/scdr/SamplerState.h>
#include <gtf/scdr/Viewport.h>
#include <gtf/scdr/WidgetRenderer.h>

#include <gtf/utils/math/matrix4x4.h>
#include <gtf/utils/smartptr/RefCountedPtr.h>

#include "CustomEffectDescriptor.h"
#include "CustomEffectGraphicObjects.h"
#include "CustomEffectRenderObject.h"

struct CustomEffectFeatures
{
    gtf::scdr::WidgetFeaturesCommon2D common2d;
};

struct CustomEffectProperties
{
    gtf::scdr::WidgetPropertiesCommon2D common2d;
};

struct CustomEffectUpdateItem
{
    CustomEffectUpdateItem*         next;
    CustomEffectRenderObject*       object;
    gtf::scdr::ImageResourceHandle* images;
};

class CustomEffectRenderer : public gtf::scdr::WidgetRenderer
{
protected:
    static const uint32_t BuiltInConstantBufferVariableCount  = 3;
    static const uint32_t BuiltInConstantBufferCount          = 1;
    static const uint32_t BuiltInConstantBufferSlot           = 0;
    static const uint32_t BuiltInFramebufferResourceSlot      = 1;
    static const uint32_t BuiltInViewProjMatOffset            = 0;
    static const uint32_t BuiltInColorationOffset             = 16 * sizeof(float);
    static const uint32_t BuiltInSizeOffset                   = 20 * sizeof(float);
    static const uint32_t BuiltInConstantBufferVariableOffset = 24 * sizeof(float);

    static const char* SCDRViewProjMat;
    static const char* SCDRColoration;
    static const char* SCDRSize;
    static const char* SCDRFramebuffer;
    static const char* SCDRFramebufferSamplerDX11;

public:
    static gtf::scdr::WidgetRenderer* createInstance (
        const gtf::viewfactory::ViewFactory& factory_,
        gtf::scdr::GraphicsContext& context_,
        const void* createParams_);

    static void destroyInstance(
        gtf::scdr::WidgetRenderer* renderer_,
        gtf::scdr::GraphicsContext& context_);

protected:
    CustomEffectRenderer(const gtf::viewfactory::ViewFactory& factory_, const CustomEffectCreateParams& desc_);

protected: // API specific methods
    virtual void setResources(gtf::scdr::ResourceBindingDescriptor* desc_, uint32_t textureCount_, uint32_t imageCount_) = 0;
    virtual const char* getVertexShader() const = 0;
    virtual const char* getPixelShader() const = 0;
    virtual uint32_t getSamplerCount(uint32_t textureCount_) const = 0;

public:
    virtual ~CustomEffectRenderer();

public: // WidgetRenderer interface
    virtual bool init(gtf::scdr::GraphicsContext& context_);
    virtual void deinit(gtf::scdr::GraphicsContext& context_);

    virtual gtf::scdr::RenderObject* createRenderObject(const gtf::widgetset::WidgetModel& model_, const gtf::coreruntime::CoreContext& context_, gtf::scdr::GraphicsContext& graphicsContext_);
    virtual void destroyRenderObject(gtf::scdr::RenderObject* object_, gtf::scdr::GraphicsContext& graphicsContext_);
    virtual bool update(gtf::scdr::GraphicsContext& context_);

    virtual bool traverse(
        gtf::widgetset::WidgetModel& widget_,
        const gtf::coreruntime::CoreContext& widgetContext_,
        const gtf::scdr::RenderObjectStackFrame& parentStackFrame_,
        gtf::scdr::RenderObjectStackFrame& objectStackFrame_,
        gtf::scdr::RenderObject& object_,
        gtf::scdr::TraverserContext& traverserContext_,
        gtf::scdr::GraphicsContext& graphicsContext_);

    virtual bool begin(gtf::scdr::GraphicsContext& context_);
    virtual bool end(gtf::scdr::GraphicsContext& context_);

private: // member functions
    void loadScalarUniforms(gtf::widgetset::WidgetModel& model_, const gtf::coreruntime::CoreContext& modelContext_, CustomEffectRenderObject& object_);
    bool loadImages(gtf::widgetset::WidgetModel& model_, const gtf::coreruntime::CoreContext& modelContext_, CustomEffectRenderObject& object_, gtf::scdr::Allocator& frameAllocator_);

    bool initShaderProgram(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram*& program_);
    void deinitShaderProgram(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram* program_);

    bool initPipelineStateObject(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram* program_);
    void deinitPipelineStateObject(gtf::scdr::GraphicsContext& context_);

    bool initResourceBindingTable(gtf::scdr::GraphicsContext& context_, gtf::scdr::ShaderProgram* program_, uint32_t& requiredCBSizeInBytes_);
    void deinitResourceBindingTable(gtf::scdr::GraphicsContext& context_);

    bool initConstantBuffer(gtf::scdr::GraphicsContext& context_, uint32_t sizeInBytes_);
    void deinitConstantBuffer(gtf::scdr::GraphicsContext& context_);

    bool initVertexBuffer(gtf::scdr::GraphicsContext& context_);
    void deinitVertexBuffer(gtf::scdr::GraphicsContext& context_);

    bool initSamplerState(gtf::scdr::GraphicsContext& context_);
    void deinitSamplerState(gtf::scdr::GraphicsContext& context_);

    CustomEffectRenderer(const CustomEffectRenderer&);
    CustomEffectRenderer& operator=(CustomEffectRenderer& other_);

protected:
    CustomEffectGraphicObjects      graphicObjects;
    const gtf::viewfactory::ViewFactory&           factory;
    CustomEffectRenderObject*       objects;
    const CustomEffectCreateParams& params;
    uint32_t resourceSlotCount;

    struct
    {
        ImageVariable*         image;
        ScalarUniformVariable* scalar;

        uint16_t imageCount;
        uint16_t scalarCount;
    } variables;

    struct
    {
        uint32_t firstSamplerResourceSlot;
        uint32_t constantBufferSize;
    } builtin;

private:
    CustomEffectUpdateItem* updates;
    CustomEffectProperties  properties;
    CustomEffectFeatures    features;
};

#endif // CUSTOM_EFFECT_RENDERER_H_INCLUDED
