////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <gtf/widgetset/WidgetModel.h>
#include <gtf/widgetset/WidgetFeatureModel.h>
#include <gtf/functor/FunctorBind.h>
#include "FocusedWidgetFeature.h"

gtf::widgetset::AbstractWidgetFeature*
createFocusRectColorInstance(gtf::widgetset::WidgetFeatureModel& model,
gtf::widgetset::WidgetContext const& ctx)
{
    return new FocusRectColorFeature(model, ctx);
}

void
destroyFocusRectColorInstance(gtf::widgetset::AbstractWidgetFeature* widget)
{
    delete static_cast<FocusRectColorFeature*>(widget);
}

FocusRectColorFeature::FocusRectColorFeature
(gtf::widgetset::WidgetFeatureModel& model, gtf::widgetset::WidgetContext const& ctx_)
: isInitialized(false)
, referencesResolved(false)
, fillColor(0)
, featureModel(model)
, ctx(ctx_)
{
    // Define the property references that are used in the implementation.  The
    // feature uses a property from another feature, namely property focused of
    // StateFocused.
    featurePropertyReferences[FeaturePropertyReferences::Focused].featureTypeRef.name = "StateFocused";
    featurePropertyReferences[FeaturePropertyReferences::Focused].featurePropertyRef.name = "focused";

    // the feature uses the fillColor property of the rectangle widget
    widgetPropertyReferences[PropertyReferences::FillColor].widgetTypeRef.name = "GtfRectangle";
    widgetPropertyReferences[PropertyReferences::FillColor].widgetPropertyRef.name = "fillColor";

    featurePropertyReferences[FeaturePropertyReferences::FeaturePropFocusedColor].featureTypeRef.name =
        "GtfFocusRectColorFeature";
    featurePropertyReferences[FeaturePropertyReferences::FeaturePropFocusedColor].featurePropertyRef.name =
        "focusedColor";
}

FocusRectColorFeature::~FocusRectColorFeature()
{ // nothing to do
}

void FocusRectColorFeature::init(gtf::widgetset::WidgetContext const& context)
{
    if (isInitialized)
    {
        return;
    }

    referencesResolved = context.getViewFactory()->resolveProperties
        (widgetPropertyReferences, PropertyReferences::WidgetPropMAX);
    referencesResolved &= context.getViewFactory()->resolveFeatureProperties
        (featurePropertyReferences, FeaturePropertyReferences::FeaturePropMAX);

    if (!referencesResolved)
    {
        // failed to resolve property references
        return;
    }

    // successfully resolved property references
    gtf::widgetset::WidgetModel *pWidgetModel = featureModel.getWidgetModel();

    // read rectangle fill color property
    pWidgetModel->scalar_readValue(ctx.getCoreContext()
        , widgetPropertyReferences[PropertyReferences::FillColor].widgetPropertyRef.ref
        , gtf::type::TypeId::eTypeId_color
        , &fillColor
        , sizeof(fillColor));

    gtf::widgetset::WidgetFeatureModel *pFocused =
        pWidgetModel->hasFeature((uint16_t)featurePropertyReferences[FeaturePropertyReferences::Focused].featureTypeRef.ref);
    if (pFocused)
    {
        // subscribe to changes of the focus property
        pFocused->subscribe(ctx.getCoreContext()
            , (uint32_t)featurePropertyReferences[FeaturePropertyReferences::Focused].featurePropertyRef.ref
            , this
            , gtf_bind(&FocusRectColorFeature::handleUpdate, this, &featureModel,
            pWidgetModel));
        // execute handleUpdate once for initialization
        handleUpdate(&featureModel, pWidgetModel);
    }

    isInitialized = true;
}

void FocusRectColorFeature::handleUpdate(gtf::widgetset::WidgetFeatureModel* focusRect,
    gtf::widgetset::WidgetModel* widgetModel)
{
    int32_t focusColor = 0;
    bool isFocused = false;

    // read focused color value
    bool success = focusRect->scalar_readValue(ctx.getCoreContext(),
        featurePropertyReferences[FeaturePropertyReferences::FeaturePropFocusedColor].featurePropertyRef.ref,
        gtf::type::TypeId::eTypeId_color, &focusColor, sizeof(focusColor));

    // resolve the focus-feature - check if the widget model has it
    gtf::widgetset::WidgetFeatureModel *focusable =
        widgetModel->hasFeature(
        (uint16_t)featurePropertyReferences[FeaturePropertyReferences::Focused].featureTypeRef.ref);

    if (focusable && ctx.getViewFactory()->isSubtype(widgetModel->getType(),
        widgetPropertyReferences[PropertyReferences::FillColor].widgetTypeRef.ref))
    {
        success = success && focusable->scalar_readValue(ctx.getCoreContext(),
            (uint16_t)featurePropertyReferences[FeaturePropertyReferences::Focused].featurePropertyRef.ref,
            gtf::type::TypeId::eTypeId_bool, &isFocused, sizeof(isFocused));

        // update fill color value, depending on the focused flag
        uint32_t newFillColor = fillColor;
        if (isFocused)
        {
            newFillColor = focusColor;
        }

        widgetModel->scalar_writeValue(ctx.getCoreContext(),
            widgetPropertyReferences[PropertyReferences::FillColor].widgetPropertyRef.ref,
            gtf::type::TypeId::eTypeId_color, &newFillColor, sizeof(newFillColor));
    }
}

void FocusRectColorFeature::prepare(gtf::widgetset::WidgetContext const & context)
{
    GTF_UNUSED_PARAM(context)
}

void FocusRectColorFeature::hide(gtf::widgetset::WidgetContext const& context)
{
    GTF_UNUSED_PARAM(context)

    if (!isInitialized || !referencesResolved)
    {
        return;
    }

    // successfully resolved property references
    gtf::widgetset::WidgetModel *pWidgetModel = featureModel.getWidgetModel();

    gtf::widgetset::WidgetFeatureModel *focused =
        pWidgetModel->hasFeature((uint16_t)featurePropertyReferences[FeaturePropertyReferences::Focused].featureTypeRef.ref);
    if (focused)
    {
        // unsubscribe from changes of the focus property
        focused->unsubscribe(ctx.getCoreContext()
            , (uint32_t)featurePropertyReferences[FeaturePropertyReferences::Focused].featurePropertyRef.ref
            , this);
    }

    isInitialized = false;
}
