////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#include <gtf/scdr/FramebufferClearFlags.h>
#include <gtf/scdr/FramebufferTextureAttachment.h>
#include <gtf/scdr/GraphicsContext.h>
#include <gtf/scdr/GraphicsExecutionContext.h>
#include <gtf/scdr/SDK.h>

#include <gtf/types/Macros.h>

#include "DirectXShader.h"
#include "GLES20Shader.h"
#include "GLES3Shader.h"
#include "RippleEffect.h"
#include "RippleEffectDesc.h"

class RippleEffectStackFrame : public gtf::scdr::RenderObjectStackFrame2D
{
public:
    virtual void load(gtf::scdr::RenderObjectStackFrame2DValues& valuesOut_) const
    {
        valuesOut_.coloration         = &coloration;
        valuesOut_.framebuffer        = framebuffer;
        valuesOut_.framebufferId      = framebufferId;
        valuesOut_.framebuffer2Screen = &framebuffer2Screen;
        valuesOut_.scissor            = &scissor;
        valuesOut_.viewport           = &viewport;
        valuesOut_.viewProjMat        = &viewProjMat;
        valuesOut_.widget2Framebuffer = &widget2Framebuffer;
        valuesOut_.depth              = depth;
    }

public:
    gtf::scdr::Viewport     viewport;
    gtf::math::float4x4     viewProjMat;
    gtf::math::float4x4     framebuffer2Screen;
    gtf::math::float4x4     widget2Framebuffer;
    gtf::scdr::Scissor      scissor;
    gtf::scdr::Color        coloration;
    gtf::scdr::Framebuffer* framebuffer;
    uint32_t                framebufferId;
    uint32_t*               depth;
};

class RippleEffectRenderObject
    : public gtf::scdr::RenderObject
    , public gtf::scdr::TouchObject
{
public:
    RippleEffectRenderObject(const gtf::widgetset::WidgetModel& model_, const gtf::scdr::WidgetFeaturesCommon2D& features_)
        : next(NULL)
        , offscreenFramebuffer(NULL)
    {
        gtf::scdr::GetWidgetFeaturesCommonModels2D(model_, features_, features);
    }

    struct StackFrameDepthPair
    {
        RippleEffectStackFrame frame;
        uint32_t               depth;
    };

    gtf::scdr::RenderObjectStackFrame* createStackFrame(gtf::scdr::RenderObjectStackFrame& parent_, gtf::scdr::Allocator& allocator_)
    {
        GTF_UNUSED_PARAM(parent_);
        StackFrameDepthPair* const pair = ALLOC_OBJECT(allocator_, StackFrameDepthPair)();

        if (pair)
        {
            pair->frame.depth = &pair->depth;
            return &pair->frame;
        }
        else
        {
            return NULL;
        }
    }

    void destroyStackFrame(gtf::scdr::RenderObjectStackFrame* frame_, gtf::scdr::Allocator& allocator_)
    {
        void* const rippleEffectFrame = static_cast<void*>(frame_);

        if (rippleEffectFrame != NULL)
        {
            StackFrameDepthPair* pair = static_cast<StackFrameDepthPair*>(rippleEffectFrame);
            FREE_OBJECT(allocator_, pair);
        }
    }

    const gtf::scdr::TouchObject* getTouchObject()
    {
        if (features.touch)
        {
            return this;
        }
        return NULL;
    }

    bool transformCoordinates(gtf::scdr::TransformMode::Value mode_, uint32_t coordCount_, gtf::scdr::TransformCoordinate* coordInOut_) const
    {
        switch (mode_)
        {
        case gtf::scdr::TransformMode::Widget2Screen:
            for (uint32_t i = 0; i < coordCount_; ++i)
            {
                gtf::scdr::CalculateCoordinateTransformation(widget2Screen, coordInOut_[i]);
            }
            return true;


        case gtf::scdr::TransformMode::Screen2Widget:
        {
            gtf::math::float4x4 invT;
            if (widget2Screen.inverse(invT))
            {
                for (uint32_t i = 0; i < coordCount_; ++i)
                {
                    gtf::scdr::CalculateCoordinateTransformation(invT, coordInOut_[i]);
                }

                return true;
            }
            return false;
        }

        case gtf::scdr::TransformMode::Unit2Screen:
            for (uint32_t i = 0; i < coordCount_; ++i)
            {
                gtf::scdr::CalculateCoordinateTransformation(unit2Screen, coordInOut_[i]);
            }
            return true;

        case gtf::scdr::TransformMode::Screen2Unit:
        {
            gtf::math::float4x4 invT;
            if (unit2Screen.inverse(invT))
            {
                for (uint32_t i = 0; i < coordCount_; ++i)
                {
                    gtf::scdr::CalculateCoordinateTransformation(invT, coordInOut_[i]);
                }
                return true;
            }
            return false;
        }
        default:
            return false;
        }
    }

    virtual void getBoundingRectangle(gtf::scdr::Rectangle<int16_t>& rectOut_) const
    {
        gtf::scdr::CalculateBoundingRectangle(*this, rectOut_);
    }

    virtual bool intersect(const gtf::scdr::ScreenPosition& pos_) const
    {
        return gtf::scdr::IntersectsBoundingRectangle(*this, pos_);
    }

public:
    RippleEffectRenderObject*               next;
    gtf::scdr::Scissor                      scissor;
    gtf::scdr::Color                        coloration;
    gtf::scdr::Viewport                     viewport;
    gtf::scdr::Framebuffer*                 framebuffer;
    gtf::scdr::Framebuffer*                 offscreenFramebuffer;
    uint32_t                                framebufferId;
    uint32_t                                offscreenFramebufferId;
    uint32_t                                drawId;
    gtf::math::float4x4                     widget2Screen;
    gtf::math::float4x4                     unit2Screen;
    RippleEffectUniforms                    uniforms;
    gtf::scdr::WidgetFeaturesCommon2DModels features;
};

struct RippleVertex
{
    float posX;
    float posY;
    float texU;
    float texV;
};

struct RenderFunctionContext
{
    RippleEffectGraphicObjects* graphicObjects;
    RippleEffectRenderObject*   renderObject;
};


gtf::scdr::RenderObject* RippleEffect::createRenderObject(
    const gtf::widgetset::WidgetModel& widget_,
    const gtf::coreruntime::CoreContext& widgetContext_,
    gtf::scdr::GraphicsContext& graphcisContext_)
{
    GTF_UNUSED_PARAM(widgetContext_);

    gtf::scdr::Allocator& allocator = graphcisContext_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    return ALLOC_OBJECT(allocator, RippleEffectRenderObject)(widget_, features.common2d);
}

void RippleEffect::destroyRenderObject(
    gtf::scdr::RenderObject* object_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    gtf::scdr::Framebuffer* fb = static_cast<RippleEffectRenderObject*>(object_)->offscreenFramebuffer;
    if (fb)
    {
        graphicsContext_.destroyFramebuffer(fb);
    }

    gtf::scdr::Allocator& allocator = graphicsContext_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    FREE_OBJECT(allocator, static_cast<RippleEffectRenderObject*>(object_));
}

gtf::scdr::WidgetRenderer* RippleEffect::createInstance (
    const gtf::viewfactory::ViewFactory& factory_,
    gtf::scdr::GraphicsContext& context_,
    const void* createParams_)
{
    const gtf::scdr::SDKHandle& handle = *static_cast<const gtf::scdr::SDKHandle*>(createParams_);
    gtf::scdr::Allocator& allocator    = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);

    return ALLOC_OBJECT(allocator, RippleEffect)(factory_, handle->getGraphicsApi());
}

void RippleEffect::destroyInstance(
    gtf::scdr::WidgetRenderer* renderer_,
    gtf::scdr::GraphicsContext& context_)
{
    gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::DefaultAllocator);
    return FREE_OBJECT(allocator, static_cast<RippleEffect*>(renderer_));
}

RippleEffect::RippleEffect(const gtf::viewfactory::ViewFactory& factory_, gtf::scdr::GraphicsApi::Value api_)
    : gtf::scdr::WidgetRenderer()
    , objects(NULL)
    , api(api_)
{
    gtf::scdr::ResolvePropertiesCommon2D(factory_, RippleEffectDesc::WidgetName, properties.common2d);
    gtf::scdr::ResolveFeaturesCommon2D(factory_, features.common2d);

    gtf::viewfactory::ViewFactory::WidgetPropertyReference wpr[] =
    {
        { { RippleEffectDesc::WidgetName, gtf::types::limits::uint16_max }, { "animationTime", gtf::types::limits::uint16_max } },
        { { RippleEffectDesc::WidgetName, gtf::types::limits::uint16_max }, { "enabled"      , gtf::types::limits::uint16_max } }
    };

    factory_.resolveProperties(wpr, ARRAY_SIZE(wpr));

    properties.animationTime = wpr[0].widgetPropertyRef.ref;
    properties.enabled       = wpr[1].widgetPropertyRef.ref;


    memset(&graphicObjects, 0, sizeof(graphicObjects));
}

RippleEffect::~RippleEffect()
{
}

static gtf::scdr::ProgramDescriptor GetProgramDescriptor(gtf::scdr::GraphicsApi::Value api_)
{
    gtf::scdr::ProgramDescriptor pd;
    memset(&pd, 0, sizeof(pd));

    switch (api_)
    {
    case gtf::scdr::GraphicsApi::OpenGLES20:
        pd.vs.source = VertexShaderGLES20;
        pd.ps.source = FragmentShaderGLES20;
        break;
    case gtf::scdr::GraphicsApi::OpenGLES3:
        pd.vs.source = VertexShaderGLES3;
        pd.ps.source = FragmentShaderGLES3;
        break;
    case gtf::scdr::GraphicsApi::DirectX11:
        pd.vs.source = VertexShaderD3D11;
        pd.ps.source = PixelShaderD3D11;
        break;

    default:
        break;
    }

    pd.ps.sourceLength = strlen(pd.ps.source);
    pd.vs.sourceLength = strlen(pd.vs.source);

    return pd;
}

static gtf::scdr::ResourceBindingTableDescriptor GetResourceBindingTableDescriptor(
    gtf::scdr::GraphicsApi::Value api_,
    gtf::scdr::ShaderProgram* program_)
{
    static const gtf::scdr::ResourceBindingDescriptor rbdGles2[] =
    {
        { "uProjMat"     , gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(RippleEffectResources, constantBuffer), offsetof(RippleEffectUniforms, projMat     ) },
        { "uModelMat"    , gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(RippleEffectResources, constantBuffer), offsetof(RippleEffectUniforms, modelMat    ) },
        { "uRippleTime"  , gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(RippleEffectResources, constantBuffer), offsetof(RippleEffectUniforms, rippleTime  ) },
        { "uRippleEnable", gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(RippleEffectResources, constantBuffer), offsetof(RippleEffectUniforms, rippleEnable) },
        { "u_texture0"   , gtf::scdr::ResourceType::Sampler       , RESOURCE_BINDING_SLOT(RippleEffectResources, sampler)       , 0                                            },
        { "u_texture0"   , gtf::scdr::ResourceType::Texture2D     , RESOURCE_BINDING_SLOT(RippleEffectResources, texture)       , 0                                            }
    };

    static const gtf::scdr::ResourceBindingDescriptor rbdGles3[] =
    {
        { "RippleParameters", gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(RippleEffectResources, constantBuffer), 0 },
        { "u_texture0"      , gtf::scdr::ResourceType::Sampler       , RESOURCE_BINDING_SLOT(RippleEffectResources, sampler)       , 0 },
        { "u_texture0"      , gtf::scdr::ResourceType::Texture2D     , RESOURCE_BINDING_SLOT(RippleEffectResources, texture)       , 0 }
    };

    static const gtf::scdr::ResourceBindingDescriptor rbdDx11[] =
    {
        { "RippleParameters", gtf::scdr::ResourceType::ConstantBuffer, RESOURCE_BINDING_SLOT(RippleEffectResources, constantBuffer), 0 },
        { "texSampler"      , gtf::scdr::ResourceType::Sampler       , RESOURCE_BINDING_SLOT(RippleEffectResources, sampler)       , 0 },
        { "tex"             , gtf::scdr::ResourceType::Texture2D     , RESOURCE_BINDING_SLOT(RippleEffectResources, texture)       , 0 }
    };

    gtf::scdr::ResourceBindingTableDescriptor rbtd = { program_, NULL, 0 };
    switch (api_)
    {
    case gtf::scdr::GraphicsApi::OpenGLES20:
        rbtd.bindings      = rbdGles2;
        rbtd.bindingsCount = ARRAY_SIZE(rbdGles2);
        return rbtd;

    case gtf::scdr::GraphicsApi::OpenGLES3:
        rbtd.bindings      = rbdGles3;
        rbtd.bindingsCount = ARRAY_SIZE(rbdGles3);
        return rbtd;

    case gtf::scdr::GraphicsApi::DirectX11:
        rbtd.bindings      = rbdDx11;
        rbtd.bindingsCount = ARRAY_SIZE(rbdDx11);
        return rbtd;

    default:
        return rbtd;
    }
}

bool RippleEffect::init(gtf::scdr::GraphicsContext& context_)
{
    const gtf::scdr::ProgramDescriptor progDesc = GetProgramDescriptor(api);
    gtf::scdr::ShaderProgram* program           = context_.createShaderProgram(progDesc);

    if (program == NULL)
    {
        return false;
    }

    // Create a pipeline state object.
    static const gtf::scdr::PipelineInputDescriptor inputDesc[] =
    {
        { "aPosition", 0, gtf::scdr::Format::Float2, 0, 0                , gtf::scdr::InputClassification::VertexData, 0 },
        { "aTexCoord", 0, gtf::scdr::Format::Float2, 0, 2 * sizeof(float), gtf::scdr::InputClassification::VertexData, 0 }
    };

    gtf::scdr::PipelineStateObjectDescriptor psoDesc;
    psoDesc.inputDescriptors      = inputDesc;
    psoDesc.inputDescriptorsCount = ARRAY_SIZE(inputDesc);
    psoDesc.program               = program;
    psoDesc.primitiveTopology     = gtf::scdr::PrimitiveTopology::TriangleStrip;

    gtf::scdr::CreateDefaultDepthStencilStateDescriptor(psoDesc.depthStencilStateDescriptor);
    gtf::scdr::CreateDefaultRasterizerStateDescriptor(psoDesc.rasterizerStateDescriptor);
    gtf::scdr::CreateAdditiveBlendStateDescriptor(psoDesc.blendStateDescriptor);

    psoDesc.rasterizerStateDescriptor.scissorEnable = true;

    graphicObjects.pipelineStateObject = context_.createPipelineStateObject(psoDesc);

    const gtf::scdr::ResourceBindingTableDescriptor rbtd = GetResourceBindingTableDescriptor(api, program);
    graphicObjects.resourceBindingTable                  = context_.createResourceBindingTable(rbtd);

    // Create a vertex buffer
    static RippleVertex vertices[] =
    {
        { 0.f, 0.f, 0.f, 0.f },
        { 0.f, 1.f, 0.f, 1.f },
        { 1.f, 0.f, 1.f, 0.f },
        { 1.f, 1.f, 1.f, 1.f }
    };

    const gtf::scdr::BufferDescriptor vetexBufferDesc = gtf::scdr::CBufferDescriptor(gtf::scdr::BufferUsage::Immutable, sizeof(vertices));
    graphicObjects.vertices                           = context_.createVertexBuffer(vetexBufferDesc, vertices);

    const gtf::scdr::BufferDescriptor uniformBufferDesc = gtf::scdr::CBufferDescriptor(gtf::scdr::BufferUsage::Dynamic, sizeof(RippleEffectUniforms));
    graphicObjects.uniforms                             = context_.createConstantBuffer(uniformBufferDesc, NULL);

    const gtf::scdr::SamplerStateDescriptor desc = gtf::scdr::CSamplerStateDescriptor(gtf::scdr::TextureFilterMode::Linear);
    graphicObjects.sampler                       = context_.createSamplerState(desc);

    // Clean up the local instance of the shader program.
    context_.destroyShaderProgram(program);

    return graphicObjects.pipelineStateObject != NULL &&
        graphicObjects.resourceBindingTable != NULL &&
        graphicObjects.sampler != NULL &&
        graphicObjects.uniforms != NULL &&
        graphicObjects.vertices != NULL;
}

void RippleEffect::deinit(gtf::scdr::GraphicsContext& context_)
{
    context_.destroySamplerState(graphicObjects.sampler);
    context_.destroyConstantBuffer(graphicObjects.uniforms);
    context_.destroyVertexBuffer(graphicObjects.vertices);
    context_.destroyResourceBindingTable(graphicObjects.resourceBindingTable);
    context_.destroyPipelineStateObject(graphicObjects.pipelineStateObject);
}

bool RippleEffect::begin(gtf::scdr::GraphicsContext&)
{
    // This method is called at the beginning of each frame rendered by the SCDR.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    objects = NULL;
    return true;
}

static uint32_t GetMaxDepth(const gtf::scdr::Framebuffer* fb_)
{
    GTF_OSAL_ASSERT(fb_);

    const gtf::scdr::FramebufferDescriptor& desc = fb_->getDescriptor();
    switch (desc.depthStencilFormat)
    {
    case gtf::scdr::TextureFormat::D16:   return (1U << 16);
    case gtf::scdr::TextureFormat::D32:   return (1U << 24);
    case gtf::scdr::TextureFormat::D24:   return (1U << 24);
    case gtf::scdr::TextureFormat::D24S8: return (1U << 24);
    default:                                   return 0;
    }
}

bool RippleEffect::traverse(
    gtf::widgetset::WidgetModel& widget_,
    const gtf::coreruntime::CoreContext& widgetContext_,
    const gtf::scdr::RenderObjectStackFrame& parentStackFrame_,
    gtf::scdr::RenderObjectStackFrame& objectStackFrame_,
    gtf::scdr::RenderObject& object_,
    gtf::scdr::TraverserContext& traverserContext_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    RippleEffectRenderObject& object    = static_cast<RippleEffectRenderObject&>(object_);
    RippleEffectStackFrame& objectFrame = static_cast<RippleEffectStackFrame&>(objectStackFrame_);

    gtf::scdr::WidgetPropertiesCommon2DValues common2dValues;
    if (!gtf::scdr::GetWidgetPropertiesCommon2DValues(widget_, widgetContext_, properties.common2d, common2dValues))
    {
        return false;
    }

    if (common2dValues.width <= 0 || common2dValues.height <= 0)
    {
        return false;
    }

    // Load common common parameters and values
    gtf::scdr::WidgetFeaturesCommon2DValues featureValues;
    gtf::scdr::RenderObjectStackFrame2DValues stackValues;

    gtf::scdr::GetWidgetFeaturesCommon2DValues(object.features, widgetContext_, features.common2d, featureValues);
    gtf::scdr::GetRenderObjectStackFrameValues2D(parentStackFrame_, stackValues);

    // Compute transformations
    gtf::math::float4x4 widget2Framebuffer = *stackValues.widget2Framebuffer *
        gtf::math::float4x4::Translation2D(static_cast<float>(common2dValues.x),
                                           static_cast<float>(common2dValues.y));

    gtf::math::float4x4 featureT;
    if (gtf::scdr::CalculateFeatureTransformMatrix2D(featureValues, featureT))
    {
        widget2Framebuffer *= featureT;
    }

    const uint32_t z = gtf::scdr::GetWidgetDepth(stackValues);
    bool enabled     = false;
    int32_t time     = 0;

    // Load effect specific parameters
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.enabled, enabled);
    gtf::scdr::LoadValue(widget_, widgetContext_, properties.animationTime, time);

    // Prepare the render object
    object.drawId                = z;
    object.widget2Screen         = *stackValues.framebuffer2Screen * widget2Framebuffer;
    object.unit2Screen           = object.widget2Screen * gtf::math::float4x4::Scaling2D(static_cast<float>(common2dValues.width), static_cast<float>(common2dValues.height));
    object.uniforms.rippleEnable = enabled ? 1.f : 0.f;
    object.uniforms.rippleTime   = static_cast<float>(time);
    object.uniforms.modelMat     = widget2Framebuffer * gtf::math::float4x4::Translation(0.f, 0.f, static_cast<float>(z)) * gtf::math::float4x4::Scaling2D(static_cast<float>(common2dValues.width), static_cast<float>(common2dValues.height));
    object.uniforms.projMat      = *stackValues.viewProjMat;
    object.viewport              = *stackValues.viewport;
    object.scissor               = *stackValues.scissor;
    object.coloration            = gtf::scdr::CalculateColoration(featureValues, stackValues);

    gtf::scdr::CalculateScissorRectangle(widget2Framebuffer, static_cast<float>(common2dValues.width), static_cast<float>(common2dValues.height), object.scissor);

    if (!createOrUpdateFramebuffer(common2dValues.width, common2dValues.height, stackValues.framebuffer, object.offscreenFramebuffer, graphicsContext_))
    {
        return false;
    }

    object.offscreenFramebufferId = traverserContext_.createFramebufferId();
    object.framebuffer            = stackValues.framebuffer;
    object.framebufferId          = stackValues.framebufferId;

    // Prepare the stack frame for child widgets
    const uint32_t maxDepth = GetMaxDepth(object.offscreenFramebuffer);
    if (maxDepth == 0)
    {
        return false;
    }

    gtf::math::float4x4 offscreenViewProjMat;
    if (!gtf::scdr::GetViewProjectionMatrix2D(common2dValues.width, common2dValues.height, maxDepth, api, offscreenViewProjMat))
    {
        return false;
    }

    objectFrame.viewport           = gtf::scdr::CViewport(0, 0, common2dValues.width, common2dValues.height);
    objectFrame.scissor            = gtf::scdr::CScissor(0, gtf::types::limits::int16_max, 0, gtf::types::limits::int16_max);
    objectFrame.framebuffer        = object.offscreenFramebuffer;
    objectFrame.framebufferId      = object.offscreenFramebufferId;
    objectFrame.viewProjMat        = offscreenViewProjMat;
    objectFrame.widget2Framebuffer = gtf::math::float4x4::Identity();
    objectFrame.framebuffer2Screen = object.widget2Screen;
    objectFrame.coloration         = object.coloration;
    *objectFrame.depth             = maxDepth - 1;

    if (featureValues.mask & gtf::scdr::WidgetFeaturesCommon2DValues::Viewport)
    {
        objectFrame.widget2Framebuffer = gtf::math::float4x4::Translation2D(
            static_cast<float>(featureValues.viewportX),
            static_cast<float>(featureValues.viewportY));
    }

    // queue the render object for drawing
    object.next = objects;
    objects     = &object;

    return true;
}

bool RippleEffect::update(gtf::scdr::GraphicsContext&)
{
    // Here is typically code issued which updates graphics resources like
    // VBOs or textures. Graphics code for updating this resources would be issued
    // with a GraphicsContext::executeUpdateFunction command.
    // Nothing to do for the ripple effect widget here.
    return true;
}

static bool DeleteRenderFunctionContext(void* memory_, gtf::scdr::GraphicsExecutionContext& context_)
{
    FREE_OBJECT(context_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator), static_cast<RenderFunctionContext*>(memory_));
    return true;
}

static bool ClearOffscreenFramebuffer(void* object_, gtf::scdr::GraphicsExecutionContext& context_)
{
    static const float clearColor[4] = { 0.0, 0.0, 0.0, 0.0 };

    if (RippleEffectRenderObject* object = static_cast<RippleEffectRenderObject*>(object_))
    {
        const uint32_t clearFlags = gtf::scdr::FramebufferClearFlags::Color
            | gtf::scdr::FramebufferClearFlags::Depth
            | gtf::scdr::FramebufferClearFlags::Stencil;

        context_.clearFramebuffer(object->offscreenFramebuffer, clearFlags, clearColor, 1.0, 0);

        return true;
    }
    return false;
}

static bool DrawEffect(void* renderFunctionContext_, gtf::scdr::GraphicsExecutionContext& context_)
{
    if (const RenderFunctionContext* const renderFunctionContext = static_cast<RenderFunctionContext*>(renderFunctionContext_))
    {
        const RippleEffectGraphicObjects* graphicObjects = renderFunctionContext->graphicObjects;
        const RippleEffectRenderObject* renderObject     = renderFunctionContext->renderObject;

        //
        // This call is required for MSAA offscreen framebuffers. It has no effect in case the framebuffer is not multisampled
        //
        context_.resolveFramebuffer(renderObject->offscreenFramebuffer);

        // Assemble resources
        gtf::scdr::Texture2D* colorTexture = context_.getTextureAttachment(
            renderObject->offscreenFramebuffer, gtf::scdr::FramebufferTextureAttachment::Color0);

        RippleEffectResources resources;
        gtf::scdr::SetResourceSlot(resources.constantBuffer, graphicObjects->uniforms);
        gtf::scdr::SetResourceSlot(resources.sampler, graphicObjects->sampler);
        gtf::scdr::SetResourceSlot(resources.texture, colorTexture);

        // Setup the GPU pipeline
        uint32_t stride = sizeof(RippleVertex);
        uint32_t offset = 0;

        // RippleEffectRenderObject specific parameters
        context_.updateConstantBuffer(graphicObjects->uniforms, &(renderObject->uniforms), sizeof(RippleEffectUniforms));

        context_.setPipelineStateObject(graphicObjects->pipelineStateObject);
        context_.setFramebuffer(renderObject->framebuffer);
        context_.setVertexBuffers(1, &(graphicObjects->vertices), &stride, &offset);
        context_.setPipelineResources(graphicObjects->resourceBindingTable, &(resources.constantBuffer), sizeof(resources) / sizeof(resources.constantBuffer));

        context_.setViewport(renderObject->viewport);
        context_.setScissor(renderObject->scissor);

        context_.draw(4, 0);

        return true;
    }

    return false;
}

bool RippleEffect::render(gtf::scdr::GraphicsContext& context_)
{
    gtf::scdr::Allocator& allocator = context_.getAllocator(gtf::scdr::AllocatorType::FrameAllocator);

    for (RippleEffectRenderObject* object = objects; object != NULL; object = object->next)
    {
        RenderFunctionContext* renderFunctionContext = ALLOC_OBJECT(allocator, RenderFunctionContext)();
        if (renderFunctionContext)
        {
            //
            // DeleteRenderFunctionContext will be executed after ClearOffscreenFramebuffer and DrawEffect
            // have been executed.
            //
            if (context_.executeDeleteFunction(DeleteRenderFunctionContext, renderFunctionContext))
            {
                renderFunctionContext->graphicObjects = &graphicObjects;
                renderFunctionContext->renderObject   = object;

                gtf::scdr::ScheduleDescriptor scheduleClear =
                    gtf::scdr::CScheduleDescriptor(object->offscreenFramebufferId, true);

                context_.executeRenderFunction(ClearOffscreenFramebuffer, object, scheduleClear);

                const gtf::scdr::CScheduleDescriptor scheduleRender(
                    graphicObjects.pipelineStateObject,
                    object->framebufferId,
                    object->drawId,
                    false);

                context_.executeRenderFunction(DrawEffect, renderFunctionContext, scheduleRender);
            }
            else
            {
                FREE_OBJECT(allocator, renderFunctionContext);
            }
        }
    }
    return true;
}

bool RippleEffect::end(gtf::scdr::GraphicsContext&)
{
    // This method is called at the end of each frame rendered.
    // Reset the internal chain of custom widget data blocks. Since the memory blocks
    // behind these pointers are managed by the SCDR there is no need for cleanup here.
    return true;
}

bool RippleEffect::createOrUpdateFramebuffer(
    int32_t width_,
    int32_t height_,
    const gtf::scdr::Framebuffer* sysFb_,
    gtf::scdr::Framebuffer*& fbInOut_,
    gtf::scdr::GraphicsContext& graphicsContext_)
{
    uint32_t width  = static_cast<uint32_t>(width_);
    uint32_t height = static_cast<uint32_t>(height_);

    // Avoid framebuffers with size 0/0
    if (0 == width_ || 0 == height_)
    {
        return false;
    }

    gtf::scdr::FramebufferDescriptor fbdesc;

    if (fbInOut_)
    {
        fbdesc = fbInOut_->getDescriptor();

        if (fbdesc.width == width && fbdesc.height == height)
        {
            return true;
        }

        graphicsContext_.destroyFramebuffer(fbInOut_);

        fbdesc.width  = width;
        fbdesc.height = height;
    }
    else
    {
        fbdesc        = sysFb_->getDescriptor();
        fbdesc.width  = width;
        fbdesc.height = height;
        fbdesc.flags  = 0;
    }

    fbInOut_ = graphicsContext_.createFramebuffer(fbdesc);

    return fbInOut_ != NULL;
}
