////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Elektrobit Automotive GmbH
// Alle Rechte vorbehalten. All Rights Reserved.
//
// Information contained herein is subject to change without notice.
// Elektrobit retains ownership and all other rights in the software and each
// component thereof.
// Any reproduction of the software or components thereof without the prior
// written permission of Elektrobit is prohibited.
////////////////////////////////////////////////////////////////////////////////

#ifndef CUSTOM_WIDGET_RIPPLE_EFFECT_H_INCLUDED
#define CUSTOM_WIDGET_RIPPLE_EFFECT_H_INCLUDED

#include <gtf/scdr/Framebuffer.h>
#include <gtf/scdr/GraphicsApi.h>
#include <gtf/scdr/GraphicsContext.h>
#include <gtf/scdr/GraphicsExecutionContext.h>
#include <gtf/scdr/RenderObjectUtils.h>
#include <gtf/scdr/SamplerState.h>
#include <gtf/scdr/Viewport.h>
#include <gtf/scdr/WidgetRenderer.h>

#include <gtf/utils/math/matrix4x4.h>

class RippleEffectRenderObject;

struct RippleEffectResources
{
    gtf::scdr::ResourceBindingSlot constantBuffer;
    gtf::scdr::ResourceBindingSlot texture;
    gtf::scdr::ResourceBindingSlot sampler;
};

struct RippleEffectProperties
{
    gtf::scdr::WidgetPropertiesCommon2D common2d;
    uint16_t                            enabled;
    uint16_t                            animationTime;
};

struct RippleEffectFeatures
{
    gtf::scdr::WidgetFeaturesCommon2D common2d;
};

struct RippleEffectUniforms
{
    gtf::math::float4x4 projMat;
    gtf::math::float4x4 modelMat;
    float               rippleTime;
    float               rippleEnable;
    float               pad[2]; // For DirectX11 to be 16 byte aligned
};

struct RippleEffectGraphicObjects
{
    gtf::scdr::ResourceBindingTable* resourceBindingTable;
    gtf::scdr::PipelineStateObject*  pipelineStateObject;
    gtf::scdr::ConstantBuffer*       uniforms;
    gtf::scdr::VertexBuffer*         vertices;
    gtf::scdr::SamplerState*         sampler;
};

//!
//! \brief Implements an effect that distorts an offscreen framebuffer with a wave pattern.
//!
class RippleEffect : public gtf::scdr::WidgetRenderer
{
private:
    RippleEffect(const RippleEffect&);
    RippleEffect& operator=(const RippleEffect&);
    RippleEffect(const gtf::viewfactory::ViewFactory& factory_, gtf::scdr::GraphicsApi::Value api_);

public:
    //!
    //! \brief Creates an instance of the RippleEffect renderer.
    //!
    //! \param[in] context_ The SCDR graphics context.
    //! \param[in] createParams_ A pointer to user specific parameters.
    //!
    //! \return An instance of the custom renderer.
    //!
    //! \note
    //!   The lifetime of the createParams_ must extend over the entire lifetime of the program. Also, the
    //!   instance will be destroyed through destroyInstance on shutdown.
    //!
    static gtf::scdr::WidgetRenderer* createInstance (
        const gtf::viewfactory::ViewFactory& factory_,
        gtf::scdr::GraphicsContext& context_,
        const void* createParams_);

    //!
    //! \brief Destroys an instance of the RippleEffect
    //!
    //! \param[in] renderer_ The renderer to destroy.
    //! \param[in] context_ The SCDR graphics context.
    //!
    //!
    static void destroyInstance(
        gtf::scdr::WidgetRenderer* renderer_,
        gtf::scdr::GraphicsContext& context_);

    virtual ~RippleEffect();

public: // gtf::scdr::WidgetRenderer interface
    virtual bool init(gtf::scdr::GraphicsContext& context_);
    virtual void deinit(gtf::scdr::GraphicsContext& context_);

    virtual bool update(gtf::scdr::GraphicsContext& context_);
    virtual bool render(gtf::scdr::GraphicsContext& context_);

    virtual bool begin(gtf::scdr::GraphicsContext& context_);
    virtual bool end(gtf::scdr::GraphicsContext& context_);

    virtual gtf::scdr::RenderObject* createRenderObject(
        const gtf::widgetset::WidgetModel& widget_,
        const gtf::coreruntime::CoreContext& widgetContext_,
        gtf::scdr::GraphicsContext& graphcisContext_);

    virtual void destroyRenderObject(
        gtf::scdr::RenderObject* object_,
        gtf::scdr::GraphicsContext& graphicsContext_);

    virtual bool traverse(
        gtf::widgetset::WidgetModel& widget_,
        const gtf::coreruntime::CoreContext& widgetContext_,
        const gtf::scdr::RenderObjectStackFrame& parentStackFrame_,
        gtf::scdr::RenderObjectStackFrame& objectStackFrame_,
        gtf::scdr::RenderObject& object_,
        gtf::scdr::TraverserContext& traverserContext_,
        gtf::scdr::GraphicsContext& graphicsContext_);

private:
    static bool createOrUpdateFramebuffer(
        int32_t width_,
        int32_t height_,
        const gtf::scdr::Framebuffer* systemFramebuffer_,
        gtf::scdr::Framebuffer*& fbInOut_,
        gtf::scdr::GraphicsContext& graphicsContext_);

private:
    // A block of graphics objects that are common for all instances of the RippleEffectRenderObject
    RippleEffectGraphicObjects graphicObjects;

    // A linked list of all RippleEffectRenderObjects to draw
    RippleEffectRenderObject* objects;

    const gtf::scdr::GraphicsApi::Value api;
    RippleEffectProperties              properties;
    RippleEffectFeatures                features;
};

#endif // CUSTOM_WIDGET_RIPPLE_EFFECT_H_INCLUDED
